/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.c
  * @brief          : Main program body
  ******************************************************************************
  ** This notice applies to any and all portions of this file
  * that are not between comment pairs USER CODE BEGIN and
  * USER CODE END. Other portions of this file, whether 
  * inserted by the user or by software development tools
  * are owned by their respective copyright owners.
  *
  * COPYRIGHT(c) 2020 STMicroelectronics
  *
  * Redistribution and use in source and binary forms, with or without modification,
  * are permitted provided that the following conditions are met:
  *   1. Redistributions of source code must retain the above copyright notice,
  *      this list of conditions and the following disclaimer.
  *   2. Redistributions in binary form must reproduce the above copyright notice,
  *      this list of conditions and the following disclaimer in the documentation
  *      and/or other materials provided with the distribution.
  *   3. Neither the name of STMicroelectronics nor the names of its contributors
  *      may be used to endorse or promote products derived from this software
  *      without specific prior written permission.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************
  */
/* USER CODE END Header */

/* Includes ------------------------------------------------------------------*/
#include "main.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
#include <string.h>
#include <stdbool.h>
#include <stdlib.h>
#include "si5351.h"
#include "lcd.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */

/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
ADC_HandleTypeDef hadc1;

I2C_HandleTypeDef hi2c1;

TIM_HandleTypeDef htim1;

/* USER CODE BEGIN PV */
/* Private variables ---------------------------------------------------------*/

/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_TIM1_Init(void);
static void MX_I2C1_Init(void);
static void MX_ADC1_Init(void);
/* USER CODE BEGIN PFP */
/* Private function prototypes -----------------------------------------------*/

/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

#define VFO_CH 0
#define BFO_CH 1
#define CAL_CH 2
#define CAL_FREQ 10000000

int32_t Fvfo = 1000000; // dummy value, will be changed below

const uint8_t eeprom_i2c_addr = (0x50 << 1);
#define CORRECTION_CONFIG_EEPROM_ADDR   0x0000
#define KEYER_CONFIG_EEPROM_ADDR        0x0020

typedef struct {
    uint16_t checksum;
    // si5351 correction factor
    int32_t correction;
} CorrectionConfig_t;

int32_t si5351_correction = 0;

// highest frequency in SSB filter passband
const int32_t BFOFrequency = 9000200;
// CW filter center frequency offset from BFO frequency
const int32_t VFOFrequencyShiftCW = -2000;

// delta to cutt off low frequencies in SSB
const int32_t BFOFrequencyShiftSSB = 300;
const int32_t VFOFrequencyShiftSSB = 300;

const int32_t CWTone = 1000;
const int32_t BFOFrequencyShiftCW = VFOFrequencyShiftCW - CWTone;

typedef enum {
    CLAR_MODE_DISABLED = 0,
    CLAR_MODE_RX = 1,
    CLAR_MODE_TX = 2,
} ClarMode_t;

ClarMode_t currentClarMode = CLAR_MODE_DISABLED;
int32_t currentClarFrequency = 0;
int32_t clarFrequencyStep = 0;
bool changingClarFrequency = false;

// used only for setting the correction factor
typedef enum {
    ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY = 0,
    ROTARY_ENCODER_MODE_CHANGE_STEP = 1,
} RotaryEncoderMode_t;

typedef enum {
    MODE_CW = 0,
    MODE_SSB = 1,
} Mode_t;

typedef struct {
    int32_t minFreq;
    int32_t maxFreq;
} BandInfo_t;

#define DEFAULT_BAND 2 // 20m
#define BANDS_NUMBER 4

BandInfo_t bandsInfo[BANDS_NUMBER] = {
    {3500000, 3800000},
    {7000000, 7200000},
    {14000000, 14350000},
    {28000000, 29700000},
};

typedef struct {
    int32_t frequency;
    int32_t frequencyStep;
    Mode_t mode;
} BandConfig_t;

typedef struct {
    uint16_t checksum;
    bool straightKey;
    int8_t speedWPM;
    int8_t ditTimeMs;
} KeyerConfig_t;

typedef struct {
    KeyerConfig_t keyer;
    int8_t currentBand;
    int32_t frequencyStep;
    BandConfig_t bands[BANDS_NUMBER];
} Config_t;

static const Config_t defaultConfig = {
    .keyer = {
        .checksum = 0,
        .straightKey = false,
        .speedWPM = 19,
        .ditTimeMs = 63,
    },
    .currentBand = DEFAULT_BAND,
    .bands = {
        { .mode = MODE_CW, .frequency =  3560000, .frequencyStep = 100 },
        { .mode = MODE_CW, .frequency =  7030000, .frequencyStep = 100 },
        { .mode = MODE_CW, .frequency = 14060000, .frequencyStep = 100 },
        { .mode = MODE_CW, .frequency = 28060000, .frequencyStep = 100 },
    }
};

// used only for setting the correction factor
RotaryEncoderMode_t rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY;

Config_t currentConfig = defaultConfig;

bool inTransmitMode = false;
uint32_t transmitModeEnterTime = 0;

#define BUTTON_DEBOUNCE_TIME_MS 150
typedef enum {
    BUTTON_PRESSED = 0,
    BUTTON_RELEASED = 1,
    BUTTON_DEBOUNCE = 2,
} ButtonStatus_t;

int32_t getVFOShift(Mode_t mode) {
    if(mode == MODE_CW) {
        return VFOFrequencyShiftCW;
    } else {
        return VFOFrequencyShiftSSB;
    }

    return 0;
}

int32_t getBFOShift(Mode_t mode) {
    if(mode == MODE_CW) {
        return BFOFrequencyShiftCW;
    } else {
        return BFOFrequencyShiftSSB;
    }

    return 0;
}

int32_t getClarRxDelta() {
    if(currentClarMode == CLAR_MODE_RX) {
        return currentClarFrequency;
    }
    return 0;
}


int32_t getClarTxDelta() {
    if(currentClarMode == CLAR_MODE_TX) {
        return currentClarFrequency;
    }
    return 0;
}

bool si5351_pllSetupDone = false;
void SetupCLK(uint8_t output, int32_t Fclk, si5351DriveStrength_t driveStrength) {
    si5351PLLConfig_t pll_conf;
    si5351OutputConfig_t out_conf;

    si5351_Calc(Fclk, &pll_conf, &out_conf);
    if(!si5351_pllSetupDone) {
        si5351_SetupPLL(SI5351_PLL_A, &pll_conf);
        si5351_pllSetupDone = true;
    }
    si5351_SetupOutput(output, SI5351_PLL_A, driveStrength, &out_conf);
}

bool inFastMode() {
    return clarFrequencyStep == 1000;
}

void displayModeAndRx() {
    LCD_Goto(0, 10);
    if(currentConfig.bands[currentConfig.currentBand].mode == MODE_CW) {
        LCD_SendString(" CW");
    } else {
        LCD_SendString("SSB");
    }
    LCD_SendString(" RX");
}

void switchFastMode() {
    if(inFastMode()) {
        currentConfig.bands[currentConfig.currentBand].frequencyStep = 100;
        clarFrequencyStep = 100;
        displayModeAndRx();
    } else {
        currentConfig.bands[currentConfig.currentBand].frequencyStep = 1000;
        clarFrequencyStep = 1000;

        LCD_Goto(0, 10);
        LCD_SendString(" FAST ");
    }
}

void changeFrequency(int32_t delta) {
    int32_t resultingFrequency;
    int32_t clarTxDelta = getClarTxDelta();
    int32_t dialFrequency = currentConfig.bands[currentConfig.currentBand].frequency;
    dialFrequency += currentConfig.bands[currentConfig.currentBand].frequencyStep*delta;

    if(dialFrequency < bandsInfo[currentConfig.currentBand].minFreq - clarTxDelta) {
        dialFrequency = bandsInfo[currentConfig.currentBand].minFreq - clarTxDelta;
    } else if(dialFrequency > bandsInfo[currentConfig.currentBand].maxFreq - clarTxDelta) {
        dialFrequency = bandsInfo[currentConfig.currentBand].maxFreq - clarTxDelta;
    }

    resultingFrequency = dialFrequency;
    if(inTransmitMode) {
        resultingFrequency += clarTxDelta;
    } else {
        resultingFrequency += getClarRxDelta();
    }

    Mode_t targetMode = currentConfig.bands[currentConfig.currentBand].mode;
    if(dialFrequency > BFOFrequency) {
        // ensures spectral inversion to get propper sideband
        Fvfo = BFOFrequency + resultingFrequency + getVFOShift(targetMode);
    } else {
        Fvfo = BFOFrequency - resultingFrequency + getVFOShift(targetMode);
    }

    SetupCLK(VFO_CH, Fvfo, SI5351_DRIVE_STRENGTH_4MA);
    currentConfig.bands[currentConfig.currentBand].frequency = dialFrequency;
}

void changeClarFrequency(int32_t delta) {
    currentClarFrequency += clarFrequencyStep*delta;
    int32_t currentFrequency = currentConfig.bands[currentConfig.currentBand].frequency;
    int32_t resultingFrequency = currentFrequency + currentClarFrequency;

    if(resultingFrequency < bandsInfo[currentConfig.currentBand].minFreq) {
        currentClarFrequency = bandsInfo[currentConfig.currentBand].minFreq - currentFrequency;
    } else if(resultingFrequency > bandsInfo[currentConfig.currentBand].maxFreq) {
        currentClarFrequency = bandsInfo[currentConfig.currentBand].maxFreq - currentFrequency;
    }

    if(currentClarFrequency > 9900) {
        currentClarFrequency = 9900;
    } else if(currentClarFrequency < -9900) {
        currentClarFrequency = -9900;
    }

    if(currentClarMode == CLAR_MODE_RX) {
        changeFrequency(0);
    }
}

void displayFrequency() {
    char buff[16];
    int32_t dialFrequency = currentConfig.bands[currentConfig.currentBand].frequency;

    snprintf(buff, sizeof(buff), "%02ld.%03ld.%02ld", 
        dialFrequency / 1000000, (dialFrequency % 1000000) / 1000, (dialFrequency % 1000) / 10);

    LCD_Goto(0, 0);
    LCD_SendString(buff);
}

void displayClarFrequency() {
    char buff[16];
    snprintf(buff, sizeof(buff), "%s%s%04d",
        (currentClarMode == CLAR_MODE_RX) ? "R" : "T",
        currentClarFrequency < 0 ? "-" : "+",
        abs(currentClarFrequency));

    LCD_Goto(1, 3);
    LCD_SendString(buff);
}

void enableTx12V() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_1, GPIO_PIN_SET);
}

void disableTx12V() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_1, GPIO_PIN_RESET);
}

void switchToSSBFilter() {
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_0, GPIO_PIN_SET);
}

void switchToCWFilter() {
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_0, GPIO_PIN_RESET);
}

void keyDown() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_13, GPIO_PIN_SET);
}

void keyUp() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_13, GPIO_PIN_RESET);
}

void switchMicAmpOn() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_14, GPIO_PIN_SET);
}

void switchMicAmpOff() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_14, GPIO_PIN_RESET);
}

void switchAfAmpOn() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_15, GPIO_PIN_SET);
}

void switchAfAmpOff() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_15, GPIO_PIN_RESET);
}

void volumeDown() {
    // partially mute Af Amp by pulling the input to the ground via a MOSFET
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_SET);
}

void volumeUp() {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0, GPIO_PIN_RESET);
}

void modeUpdated() {
    Mode_t targetMode = currentConfig.bands[currentConfig.currentBand].mode;

    LCD_Goto(0, 10);
    if(targetMode == MODE_CW) {
        LCD_SendString(" CW");
        switchToCWFilter();
    } else {
        LCD_SendString("SSB");
        switchToSSBFilter();
    }

    SetupCLK(BFO_CH, BFOFrequency + getBFOShift(targetMode), SI5351_DRIVE_STRENGTH_4MA);
}

void switchBandPassFilter(int8_t currentBand) {
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_3, currentBand == 0 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 80m
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_4, currentBand == 1 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 40m
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_5, currentBand == 2 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 20m
    HAL_GPIO_WritePin(GPIOB, GPIO_PIN_7, currentBand == 3 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 10m
}

// pass -1 to switch all relays off
void switchLowPassFilter(int8_t currentBand) {
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_10, currentBand == 0 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 80m: LPF 80
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_11, currentBand == 1 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 40m: LPF 30, 40
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_12, currentBand == 2 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 20m: LPF 20, 17
    HAL_GPIO_WritePin(GPIOA, GPIO_PIN_15, currentBand == 3 ? GPIO_PIN_SET : GPIO_PIN_RESET); // 10m: LPF 15, 12, 10
}

void updateBottomLine(bool); // implemented below

void disableClar() {
    if(currentClarMode == CLAR_MODE_DISABLED) {
        return;
    }

    currentClarMode = CLAR_MODE_DISABLED;
    currentClarFrequency = 0;
    changingClarFrequency = false;

    changeFrequency(0); // in case RIT was enabled
    updateBottomLine(true);
}

void changeBand(int32_t delta) {
    currentConfig.currentBand += delta;

    if(currentConfig.currentBand < 0) {
        currentConfig.currentBand = BANDS_NUMBER - 1;
    } else if(currentConfig.currentBand >= BANDS_NUMBER) {
        currentConfig.currentBand = 0;
    }

    switchBandPassFilter(currentConfig.currentBand);

    disableClar();
    changeFrequency(0);
    modeUpdated();
    displayFrequency();
}

void changeMode() {
    if(currentConfig.bands[currentConfig.currentBand].mode == MODE_SSB) {
        currentConfig.bands[currentConfig.currentBand].mode = MODE_CW;
    } else {
        currentConfig.bands[currentConfig.currentBand].mode = MODE_SSB;
    }

    changeFrequency(0);
    modeUpdated();
}

void ensureTransmitMode() {
    if(inTransmitMode) {
        return;
    }

    Mode_t targetMode = currentConfig.bands[currentConfig.currentBand].mode;

    LCD_Goto(0, 14);
    LCD_SendString("TX");
    
    if(targetMode == MODE_SSB) {
        // "magic" delay reduces popping sound a little
        switchAfAmpOff();
        HAL_Delay(25);
    } else {
        keyUp();
        volumeDown();
    }

    inTransmitMode = true;
    changeFrequency(0); // account for CLAR, if necessary
    switchLowPassFilter(currentConfig.currentBand);
    enableTx12V();

    if(targetMode == MODE_SSB) {
        // don't enable microphone amplifier while transient
        // processes take place to prevent parasitic oscillations
        HAL_Delay(25);
        switchMicAmpOn();
    }
}

void ensureReceiveMode() {
    if(!inTransmitMode) {
        return;
    }

    LCD_Goto(0, 14);
    LCD_SendString("RX");

    disableTx12V();
    keyUp();
    switchMicAmpOff();
    volumeUp();
    switchAfAmpOn();
    switchLowPassFilter(-1);

    inTransmitMode = false;
    changeFrequency(0); // account for CLAR, if necessary
}

ButtonStatus_t buttonRotEncoderPressed(uint32_t now) {
    static uint32_t lastPressed = 0;
    if(HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_10) == GPIO_PIN_RESET) {
        if(now - lastPressed > BUTTON_DEBOUNCE_TIME_MS) {
            lastPressed = now;
            return BUTTON_PRESSED;
        } else {
            return BUTTON_DEBOUNCE;
        }
    }
    return BUTTON_RELEASED;
}

bool buttonBandPressedSimple() {
    return (HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_3) == GPIO_PIN_RESET);
}

ButtonStatus_t buttonBandPressed(uint32_t now) {
    static uint32_t lastPressed = 0;
    if(HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_3) == GPIO_PIN_RESET) {
        if(now - lastPressed > BUTTON_DEBOUNCE_TIME_MS) {
            lastPressed = now;
            return BUTTON_PRESSED;
        } else {
            return BUTTON_DEBOUNCE;
        }
    }
    return BUTTON_RELEASED;
}

bool buttonModePressedSimple() {
    return (HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_4) == GPIO_PIN_RESET);
}

ButtonStatus_t buttonModePressed(uint32_t now) {
    static uint32_t lastPressed = 0;
    if(HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_4) == GPIO_PIN_RESET) {
        if(now - lastPressed > BUTTON_DEBOUNCE_TIME_MS) {
            lastPressed = now;
            return BUTTON_PRESSED;
        } else {
            return BUTTON_DEBOUNCE;
        }
    }
    return BUTTON_RELEASED;
}

bool buttonClarPressedSimple() {
    return (HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_5) == GPIO_PIN_RESET);
}

ButtonStatus_t buttonClarPressed(uint32_t now) {
    static uint32_t lastPressed = 0;
    if(HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_5) == GPIO_PIN_RESET) {
        if(now - lastPressed > BUTTON_DEBOUNCE_TIME_MS) {
            lastPressed = now;
            return BUTTON_PRESSED;
        } else {
            return BUTTON_DEBOUNCE;
        }
    }
    return BUTTON_RELEASED;
}

bool buttonKeyerPressedSimple() {
    return (HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_6) == GPIO_PIN_RESET);
}

ButtonStatus_t buttonKeyerPressed(uint32_t now) {
    static uint32_t lastPressed = 0;
    if(HAL_GPIO_ReadPin(GPIOA, GPIO_PIN_6) == GPIO_PIN_RESET) {
        if(now - lastPressed > BUTTON_DEBOUNCE_TIME_MS) {
            lastPressed = now;
            return BUTTON_PRESSED;
        } else {
            return BUTTON_DEBOUNCE;
        }
    }
    return BUTTON_RELEASED;
}

bool buttonDitPressed() {
    return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_11) == GPIO_PIN_RESET);
}

bool buttonDahPressed() {
    return (HAL_GPIO_ReadPin(GPIOB, GPIO_PIN_12) == GPIO_PIN_RESET);
}

int32_t sendStart = 0;
int32_t sendFinish = 0;
int32_t keyReadTimeout = 0;
int32_t sendTimeout = 0;
bool isSending = false;

typedef enum {
    CW_SEND_NONE = 0,
    CW_SEND_DIT = 1,
    CW_SEND_DAH = 2,
} CwSend_t;

CwSend_t lastSent = CW_SEND_NONE;
CwSend_t nextSend = CW_SEND_NONE;

void processStraightKeyerLogic(bool pressed) {
    if(pressed) {
        keyDown();
    } else {
        keyUp();
    }
}

void processIambicKeyerLogic(uint32_t now, bool ditPressed, bool dahPressed) {
    if(isSending && (now >= sendFinish)) {
        keyUp();
        isSending = false;
    }

    if(nextSend == CW_SEND_NONE) {
        if(ditPressed && (lastSent == CW_SEND_DAH)) {
            nextSend = CW_SEND_DIT;
        } else if(dahPressed && (lastSent == CW_SEND_DIT)) {
            nextSend = CW_SEND_DAH;
        } else if(now > keyReadTimeout) {
            if(ditPressed) {
                nextSend = CW_SEND_DIT;
            } else if (dahPressed) {
                nextSend = CW_SEND_DAH;
            }
        }
    }

    if((now > sendTimeout) && (nextSend != CW_SEND_NONE)) {
        if(nextSend == CW_SEND_DIT) {
            sendFinish = now + currentConfig.keyer.ditTimeMs;
            keyReadTimeout = now + (currentConfig.keyer.ditTimeMs)*2;
        } else if(nextSend == CW_SEND_DAH) {
            sendFinish = now + (currentConfig.keyer.ditTimeMs)*3;
            keyReadTimeout = now + (currentConfig.keyer.ditTimeMs)*3;
        }

        lastSent = nextSend;
        nextSend = CW_SEND_NONE;
        sendStart = now;
        sendTimeout = sendFinish + currentConfig.keyer.ditTimeMs;
        isSending = true;
        keyDown();
    }
}

void ADC_Select_Channel(uint32_t ch) {
    ADC_ChannelConfTypeDef conf = {
        .Channel = ch,
        .Rank = 1,
        .SamplingTime = ADC_SAMPLETIME_28CYCLES_5,
    };
    if (HAL_ADC_ConfigChannel(&hadc1, &conf) != HAL_OK) {
        Error_Handler();
    }
}

#define LOOP_DELAY 5

void updateBottomLine(bool forceUpdate) {
    static uint8_t prevSValue = 0;
    static uint8_t counter = 0;
    static double prevVoltage = 0.0;

    if((!forceUpdate) && (++counter < (uint8_t)(500/LOOP_DELAY))) {
        // update once in 500 ms
        return;
    }

    counter = 0;

    ADC_Select_Channel(ADC_CHANNEL_1);
    HAL_ADC_PollForConversion(&hadc1, HAL_MAX_DELAY);
    uint32_t voltageADCVal = HAL_ADC_GetValue(&hadc1);

    ADC_Select_Channel(ADC_CHANNEL_2);
    HAL_ADC_PollForConversion(&hadc1, HAL_MAX_DELAY);
    uint32_t sMeterADCVal = HAL_ADC_GetValue(&hadc1);

    double voltage = ((double)(voltageADCVal)-1220.0)/190.0 + 7.1;
    bool updateVoltage = forceUpdate || (abs((int)(voltage*10) - (int)(prevVoltage*10)) > 1);

    // re-render S-meter, but only if the value has changed
    // rendering S-meter all the time makes the underline blink
    if((sMeterADCVal == 0)) {
        if(forceUpdate || (prevSValue != 2)) {
            LCD_Goto(1, 1);
            LCD_SendString("0");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("       ");
            }
            prevSValue = 2;
        }
    } else if((sMeterADCVal <= 3)) {
        if(forceUpdate || (prevSValue != 3)) {
            LCD_Goto(1, 1);
            LCD_SendString("3");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("\x01      ");
            }
            prevSValue = 3;
        }
    } else if((sMeterADCVal <= 10)) {
        if(forceUpdate || (prevSValue != 4)) {
            LCD_Goto(1, 1);
            LCD_SendString("4");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("\x01\x02     ");
            }
            prevSValue = 4;
        }
    } else if((sMeterADCVal <= 150)) {
        if(forceUpdate || (prevSValue != 5)) {
            LCD_Goto(1, 1);
            LCD_SendString("5");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("\x01\x02\x03    ");
            }
            prevSValue = 5;
        }
    } else if((sMeterADCVal <= 550)) {
        if(forceUpdate || (prevSValue != 6)) {
            LCD_Goto(1, 1);
            LCD_SendString("6");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("\x01\x02\x03\x04   ");
            }
            prevSValue = 6;
        }
    } else if((sMeterADCVal <= 1100)) {
        if(forceUpdate || (prevSValue != 7)) {
            LCD_Goto(1, 1);
            LCD_SendString("7");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("\x01\x02\x03\x04\x05  ");
            }
            prevSValue = 7;
        }
    } else if((sMeterADCVal <= 1400)) {
        if(forceUpdate || (prevSValue != 8)) {
            LCD_Goto(1, 1);
            LCD_SendString("8");
            if(currentClarMode == CLAR_MODE_DISABLED) {
                LCD_SendString("\x01\x02\x03\x04\x05\x06 ");
            }
            prevSValue = 8;
        }
    } else if((sMeterADCVal <= 1900)) {
        if(forceUpdate || (prevSValue != 9)) {
            LCD_Goto(1, 1);
            LCD_SendString("9");
            if((currentClarMode == CLAR_MODE_DISABLED) && (forceUpdate || (prevSValue != 10))) {
                LCD_SendString("\x01\x02\x03\x04\x05\x06\x07");
            }
            prevSValue = 9;
        }
    } else {
        if(forceUpdate || (prevSValue != 10)) {
            LCD_Goto(1, 1);
            LCD_SendString("+");
            if((currentClarMode == CLAR_MODE_DISABLED) && (forceUpdate || (prevSValue < 9))) {
                LCD_SendString("\x01\x02\x03\x04\x05\x06\x07");
            }
            prevSValue = 10;
        }
    }

    if((!changingClarFrequency) && updateVoltage) {
        // don't forget to add '-u _printf_float' to LDFLAGS
        char buff[16];
        if(voltage < 10) {
            snprintf(buff, sizeof(buff), "  %.1fV", voltage);
        } else {
            snprintf(buff, sizeof(buff), " %.1fV", voltage);
        }

        LCD_Goto(1, 10);
        LCD_SendString(buff);
        prevVoltage = voltage;
    }

    if(forceUpdate && (currentClarMode != CLAR_MODE_DISABLED)) {
        displayClarFrequency();
    }
}

// returns 0 if rotary encoder was not rotated
int32_t getRotaryEncoderDelta() {
    static int32_t prevCounter = 0;
    int32_t currCounter = __HAL_TIM_GET_COUNTER(&htim1);
    currCounter = 32767 - ((currCounter-1) & 0xFFFF) / 2;
    if(currCounter > 32768/2) {
        // convert ... 32766, 32767, 0, 1, 2 ... into:
        //               ... -2, -1, 0, 1, 2 ...
        // this simplifies `delta` calculation
        currCounter = currCounter-32768;
    }
    if(currCounter != prevCounter) {
        int32_t delta = currCounter-prevCounter;
        prevCounter = currCounter;
        // debounce or skip counter overflow
        if((delta > -10) && (delta < 10) && (!inTransmitMode)) {
            return delta;
        }
    }

    return 0;
}

void keyerMenuDisplayWpm() {
    char buff[32];
    
    LCD_Goto(0, 0);
    
    if(currentConfig.keyer.straightKey) {
        LCD_SendString("KEYER OFF       ");
    } else {
        snprintf(buff, sizeof(buff), "SPEED: %02d WPM   ", currentConfig.keyer.speedWPM);
        LCD_SendString(buff);
        LCD_Goto(0, 8);
    }
}

// http://en.wikipedia.org/wiki/Jenkins_hash_function
uint32_t jenkinsHash(const uint8_t *data, const size_t len) {
  uint32_t hash, i;
  for(hash = i = 0; i < len; ++i) {
    hash += data[i];
    hash += (hash << 10);
    hash ^= (hash >> 6);
  }
  hash += (hash << 3);
  hash ^= (hash >> 11);
  hash += (hash << 15);
  return hash;
}

void loadCorrectionConfig() {
    HAL_StatusTypeDef status;
    uint16_t savedChecksum;
    CorrectionConfig_t savedCorrectionConfig;

    for(;;) {
        status = HAL_I2C_IsDeviceReady(&hi2c1, eeprom_i2c_addr, 1, HAL_MAX_DELAY);
        if(status == HAL_OK)
            break;
    }

    HAL_I2C_Mem_Read(&hi2c1, eeprom_i2c_addr, CORRECTION_CONFIG_EEPROM_ADDR, I2C_MEMADD_SIZE_16BIT,
        (uint8_t*)&savedCorrectionConfig, sizeof(savedCorrectionConfig), HAL_MAX_DELAY);

    savedChecksum = savedCorrectionConfig.checksum;
    savedCorrectionConfig.checksum = 0;

    if((jenkinsHash((const uint8_t*)&savedCorrectionConfig, sizeof(savedCorrectionConfig)) & 0xFFFF) == savedChecksum) {
        si5351_correction = savedCorrectionConfig.correction;
    } else {
        LCD_Goto(0, 0);
        LCD_SendString(" SI5351  CONFIG ");
        LCD_Goto(1, 0);
        LCD_SendString(" CHECKSUM WRONG ");
        HAL_Delay(2000);
        LCD_Clear();
    }
}

void loadKeyerConfig() {
    HAL_StatusTypeDef status;
    uint16_t savedChecksum;
    KeyerConfig_t savedKeyerConfig;

    for(;;) {
        status = HAL_I2C_IsDeviceReady(&hi2c1, eeprom_i2c_addr, 1, HAL_MAX_DELAY);
        if(status == HAL_OK)
            break;
    }

    HAL_I2C_Mem_Read(&hi2c1, eeprom_i2c_addr, KEYER_CONFIG_EEPROM_ADDR, I2C_MEMADD_SIZE_16BIT,
        (uint8_t*)&savedKeyerConfig, sizeof(savedKeyerConfig), HAL_MAX_DELAY);

    savedChecksum = savedKeyerConfig.checksum;
    savedKeyerConfig.checksum = 0;

    if((jenkinsHash((const uint8_t*)&savedKeyerConfig, sizeof(savedKeyerConfig)) & 0xFFFF) == savedChecksum) {
        currentConfig.keyer = savedKeyerConfig;
    } else {
        LCD_Goto(0, 0);
        LCD_SendString("  KEYER CONFIG  ");
        LCD_Goto(1, 0);
        LCD_SendString(" CHECKSUM WRONG ");
        HAL_Delay(2000);

        LCD_Clear();
    }
}

void saveCorrectionConfig() {
    HAL_StatusTypeDef status;
    uint16_t savedChecksum;
    CorrectionConfig_t savedCorrectionConfig = {
        .checksum = 0,
        .correction = si5351_correction,
    };

    savedChecksum = (jenkinsHash((const uint8_t*)&savedCorrectionConfig, sizeof(savedCorrectionConfig)) & 0xFFFF);
    savedCorrectionConfig.checksum = savedChecksum;

    for(;;) {
        status = HAL_I2C_IsDeviceReady(&hi2c1, eeprom_i2c_addr, 1, HAL_MAX_DELAY);
        if(status == HAL_OK)
            break;
    }

    HAL_I2C_Mem_Write(&hi2c1, eeprom_i2c_addr, CORRECTION_CONFIG_EEPROM_ADDR, I2C_MEMADD_SIZE_16BIT,
        (uint8_t*)&savedCorrectionConfig, sizeof(savedCorrectionConfig), HAL_MAX_DELAY);
}

void saveKeyerConfig() {
    HAL_StatusTypeDef status;
    uint16_t savedChecksum;
    KeyerConfig_t savedKeyerConfig = currentConfig.keyer;
    savedKeyerConfig.checksum = 0;

    savedChecksum = (jenkinsHash((const uint8_t*)&savedKeyerConfig, sizeof(savedKeyerConfig)) & 0xFFFF);
    savedKeyerConfig.checksum = savedChecksum;

    for(;;) {
        status = HAL_I2C_IsDeviceReady(&hi2c1, eeprom_i2c_addr, 1, HAL_MAX_DELAY);
        if(status == HAL_OK)
            break;
    }

    HAL_I2C_Mem_Write(&hi2c1, eeprom_i2c_addr, KEYER_CONFIG_EEPROM_ADDR, I2C_MEMADD_SIZE_16BIT,
        (uint8_t*)&savedKeyerConfig, sizeof(savedKeyerConfig), HAL_MAX_DELAY);
}

// TODO rewrite
void keyerMenu() {
    LCD_Goto(1, 0);
    LCD_SendString("ON  OFF      CFM");
    
    keyerMenuDisplayWpm();

    for(;;) {
        uint32_t now = HAL_GetTick();

        if(!currentConfig.keyer.straightKey) {
            int32_t delta = getRotaryEncoderDelta();
            if(delta != 0) {
                // WPM changed
                currentConfig.keyer.speedWPM += delta;
                if(currentConfig.keyer.speedWPM < 10) {
                    currentConfig.keyer.speedWPM = 10;
                } else if (currentConfig.keyer.speedWPM > 30) {
                    currentConfig.keyer.speedWPM = 30;
                }

                // 50 is the length in dits of "PARIS "
                // see https://morsecode.world/international/timing.html
                currentConfig.keyer.ditTimeMs = 60*1000/(50*currentConfig.keyer.speedWPM);
                keyerMenuDisplayWpm();
            }
        }

        if(buttonRotEncoderPressed(now) == BUTTON_PRESSED) {
            // no action in this mode
        }

        if(buttonBandPressed(now) == BUTTON_PRESSED) {
            // 'ON' pressed
            if(currentConfig.keyer.straightKey == true) {
                currentConfig.keyer.straightKey = false;
                currentConfig.keyer.ditTimeMs = 60*1000/(50*currentConfig.keyer.speedWPM);
                keyerMenuDisplayWpm();
            }
        }

        if(buttonModePressed(now) == BUTTON_PRESSED) {
            // 'OFF' pressed
            if(currentConfig.keyer.straightKey == false) {
                currentConfig.keyer.straightKey = true;
                // ~15 WPM, used for RX/TX switching delay
                currentConfig.keyer.ditTimeMs = 80; 
                keyerMenuDisplayWpm();
            }
        }

        if(buttonClarPressed(now) == BUTTON_PRESSED) {
            // do nothing yet
        }

        if(buttonKeyerPressed(now) == BUTTON_PRESSED) {
            // 'CFM' pressed
            break;
        }
        HAL_Delay(LOOP_DELAY);
    }

    saveKeyerConfig();

    LCD_Goto(0, 9);
    LCD_SendString(" ");
    Mode_t targetMode = currentConfig.bands[currentConfig.currentBand].mode;
    if(targetMode == MODE_CW) {
        LCD_SendString(" CW");
    } else {
        LCD_SendString("SSB");
    }
    LCD_SendString(" RX");

    LCD_Goto(1, 0);
    LCD_SendString("S");
    updateBottomLine(true);
    
    displayFrequency();
}

int32_t correctionFrequencyStepToDigitOffset(int32_t step) {
    switch(step) {
        case 1:
            return 4;
        case 10:
            return 3;
        case 100:
            return 2;
        default:
            return 1;
    }
}

void setCorrectionFactorMode() {
    bool correctionChanged = true;
    int32_t step = 1;
    int32_t correction = si5351_correction;
    char buff[16];

    LCD_Clear();
    LCD_UnderlineEnabled(false);
    LCD_Goto(0, 0);
    LCD_SendString("SI5351 CF:");
    LCD_Goto(1, 0);
    LCD_SendString("CFM ABRT");
    HAL_Delay(2000);
    LCD_UnderlineEnabled(true);

    for(;;) {
        if(buttonBandPressedSimple()) {
            si5351_correction = correction;
            saveCorrectionConfig();
            break;
        }

        if(buttonModePressedSimple()) {
            break;
        }

        uint32_t now = HAL_GetTick();
        if(buttonRotEncoderPressed(now) == BUTTON_PRESSED) {
            if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY) {
                rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_STEP;
            } else {
                rotaryEncoderMode = ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY;
            }
        }

        int32_t delta = getRotaryEncoderDelta();
        if(delta != 0) {
            if(rotaryEncoderMode == ROTARY_ENCODER_MODE_CHANGE_FREQEUNCY) {
                correction += delta*step;
                if(correction < -9999) {
                    correction = -9999;
                } else if(correction > 9999) {
                    correction = 9999;
                }
            } else {
                if(delta > 0) {
                    step = step / 10;
                    if(step < 1 ) {
                        step = 1000;
                    }
                } else {
                    step = step * 10;
                    if(step > 1000) {
                        step = 1;
                    }
                }
            }

            correctionChanged = true;
        }

        if(correctionChanged) {
            correctionChanged = false;
            snprintf(buff, sizeof(buff), "%s%04d", correction >= 0 ? "+" : "-", abs(correction));
            LCD_Goto(0, 11);
            LCD_SendString(buff);
            LCD_Goto(0, 11 + correctionFrequencyStepToDigitOffset(step));

            si5351_Init(correction);
            si5351_EnableOutputs(0);
            si5351_pllSetupDone = false;
            SetupCLK(CAL_CH, CAL_FREQ, SI5351_DRIVE_STRENGTH_4MA);
            si5351_EnableOutputs(1 << CAL_CH);
        }

        HAL_Delay(LOOP_DELAY);
    }

    LCD_UnderlineEnabled(false);
}

void init() {
    // make sure we will not hear CW tone generator during the initialization
    disableTx12V();
    switchAfAmpOff();
    switchMicAmpOff();
    keyUp();

    LCD_Init();
    HAL_ADC_Start(&hadc1);
    HAL_TIM_Encoder_Start(&htim1, TIM_CHANNEL_ALL);

    loadCorrectionConfig();
    loadKeyerConfig();

    // BAND + MODE pressed: set si5351 correction factor
    if(buttonBandPressedSimple() && buttonModePressedSimple()) {
        setCorrectionFactorMode();
    }

    // CLAR + KEYER pressed: perform I2C scan
    if(buttonClarPressedSimple() && buttonKeyerPressedSimple()) {
        LCD_Goto(0, 0);
        LCD_SendString("I2C Scan:");
        LCD_Goto(1, 0);
        for(uint16_t i = 0; i < 128; i++) {
            HAL_StatusTypeDef res;
            res = HAL_I2C_IsDeviceReady(&hi2c1, i << 1, 1, 10);
            if(res == HAL_OK) {
                char msg[64];
                snprintf(msg, sizeof(msg), "%02X ", i);
                LCD_SendString(msg);
            }
        }  
        HAL_Delay(3000);
        LCD_Clear();
    }

    LCD_Goto(0, 0);
    LCD_SendString(" Homebrew Radio ");
    LCD_Goto(1, 0);
    LCD_SendString("HBR-1 R2AUK 2021");
    HAL_Delay(2000);
    LCD_Clear();

    si5351_Init(si5351_correction);

    // BFO frequency never changes
    si5351_EnableOutputs((1<<VFO_CH)|(1<<BFO_CH));

    inTransmitMode = true;
    ensureReceiveMode();

    LCD_Goto(1, 0);
    LCD_SendString("S");
    changeBand(0);
  
    displayFrequency();
}

void loop() {
    bool ditPressed = buttonDitPressed();
    bool dahPressed = buttonDahPressed();
    Mode_t targetMode = currentConfig.bands[currentConfig.currentBand].mode;
    uint32_t now = HAL_GetTick();

    if((ditPressed || dahPressed) && !inFastMode() && !changingClarFrequency) {
        if(targetMode == MODE_CW) {
            transmitModeEnterTime = now;
        }
        ensureTransmitMode();
    } else {
        if(targetMode == MODE_CW) {
            if(now - transmitModeEnterTime > (currentConfig.keyer.ditTimeMs)*10) {
                ensureReceiveMode();
            }
        } else {
            ensureReceiveMode();
        }
    }

    int32_t delta = getRotaryEncoderDelta();
    if(delta != 0) {
        if(changingClarFrequency) {
            changeClarFrequency(delta);
            displayClarFrequency();
        } else {
            changeFrequency(delta);
            displayFrequency();
        }
    }

    if(inTransmitMode) {
        Mode_t targetMode = currentConfig.bands[currentConfig.currentBand].mode;
        if(targetMode == MODE_CW) {
            if(currentConfig.keyer.straightKey) {
                processStraightKeyerLogic(ditPressed);
            } else {
                processIambicKeyerLogic(now, ditPressed, dahPressed);
            }
        }
    } else {
        if(buttonRotEncoderPressed(now) == BUTTON_PRESSED) {
            if(changingClarFrequency) {
                currentClarFrequency = 0;
                LCD_Goto(1, 2);
                if(currentClarMode == CLAR_MODE_RX) {
                    currentClarMode = CLAR_MODE_TX;
                    LCD_SendString(" T+0000");
                    changeFrequency(0); // in case RIT was non-zero
                } else {
                    currentClarMode = CLAR_MODE_RX;
                    LCD_SendString(" R+0000");
                    changeFrequency(0);
                }

                changingClarFrequency = true;
                updateBottomLine(true);
            } else {
                switchFastMode();
            }
        }

        if((buttonBandPressed(now) == BUTTON_PRESSED) && !inFastMode() && !changingClarFrequency) {
            changeBand(+1);
        }

        if((buttonModePressed(now) == BUTTON_PRESSED) && !inFastMode() && !changingClarFrequency) {
            changeMode();
        }

        if((buttonClarPressed(now) == BUTTON_PRESSED) && !inFastMode()) {
            if(currentClarMode == CLAR_MODE_DISABLED) {
                currentClarMode = CLAR_MODE_RX;
                currentClarFrequency = 0;
                clarFrequencyStep = 100;
                LCD_Goto(1, 2);
                LCD_SendString(" R+0000");
            }

            changingClarFrequency = !changingClarFrequency;
            if(changingClarFrequency) {
                LCD_Goto(0, 10);
                LCD_SendString(" CLAR ");
            } else {
                displayModeAndRx();
            }
            updateBottomLine(true);
        }

        if(!changingClarFrequency && (currentClarMode != CLAR_MODE_DISABLED) && (currentClarFrequency == 0)) {
            disableClar();
        }

        if((buttonKeyerPressed(now) == BUTTON_PRESSED) && !inFastMode() && !changingClarFrequency) {
            keyerMenu();
        }

        updateBottomLine(false);
    }

    HAL_Delay(LOOP_DELAY);
}

/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_TIM1_Init();
  MX_I2C1_Init();
  MX_ADC1_Init();
  /* USER CODE BEGIN 2 */
  init();
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    loop();
    /* USER CODE END WHILE */

    /* USER CODE BEGIN 3 */

  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};
  RCC_PeriphCLKInitTypeDef PeriphClkInit = {0};

  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_NONE;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }
  /**Initializes the CPU, AHB and APB busses clocks 
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_HSI;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_0) != HAL_OK)
  {
    Error_Handler();
  }
  PeriphClkInit.PeriphClockSelection = RCC_PERIPHCLK_ADC;
  PeriphClkInit.AdcClockSelection = RCC_ADCPCLK2_DIV2;
  if (HAL_RCCEx_PeriphCLKConfig(&PeriphClkInit) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief ADC1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_ADC1_Init(void)
{

  /* USER CODE BEGIN ADC1_Init 0 */

  /* USER CODE END ADC1_Init 0 */

  ADC_ChannelConfTypeDef sConfig = {0};

  /* USER CODE BEGIN ADC1_Init 1 */

  /* USER CODE END ADC1_Init 1 */
  /**Common config 
  */
  hadc1.Instance = ADC1;
  hadc1.Init.ScanConvMode = ADC_SCAN_DISABLE;
  hadc1.Init.ContinuousConvMode = ENABLE;
  hadc1.Init.DiscontinuousConvMode = DISABLE;
  hadc1.Init.ExternalTrigConv = ADC_SOFTWARE_START;
  hadc1.Init.DataAlign = ADC_DATAALIGN_RIGHT;
  hadc1.Init.NbrOfConversion = 1;
  if (HAL_ADC_Init(&hadc1) != HAL_OK)
  {
    Error_Handler();
  }
  /**Configure Regular Channel 
  */
  sConfig.Channel = ADC_CHANNEL_1;
  sConfig.Rank = ADC_REGULAR_RANK_1;
  sConfig.SamplingTime = ADC_SAMPLETIME_1CYCLE_5;
  if (HAL_ADC_ConfigChannel(&hadc1, &sConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN ADC1_Init 2 */

  /* USER CODE END ADC1_Init 2 */

}

/**
  * @brief I2C1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_I2C1_Init(void)
{

  /* USER CODE BEGIN I2C1_Init 0 */

  /* USER CODE END I2C1_Init 0 */

  /* USER CODE BEGIN I2C1_Init 1 */

  /* USER CODE END I2C1_Init 1 */
  hi2c1.Instance = I2C1;
  hi2c1.Init.ClockSpeed = 400000;
  hi2c1.Init.DutyCycle = I2C_DUTYCYCLE_2;
  hi2c1.Init.OwnAddress1 = 0;
  hi2c1.Init.AddressingMode = I2C_ADDRESSINGMODE_7BIT;
  hi2c1.Init.DualAddressMode = I2C_DUALADDRESS_DISABLE;
  hi2c1.Init.OwnAddress2 = 0;
  hi2c1.Init.GeneralCallMode = I2C_GENERALCALL_DISABLE;
  hi2c1.Init.NoStretchMode = I2C_NOSTRETCH_DISABLE;
  if (HAL_I2C_Init(&hi2c1) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN I2C1_Init 2 */

  /* USER CODE END I2C1_Init 2 */

}

/**
  * @brief TIM1 Initialization Function
  * @param None
  * @retval None
  */
static void MX_TIM1_Init(void)
{

  /* USER CODE BEGIN TIM1_Init 0 */

  /* USER CODE END TIM1_Init 0 */

  TIM_Encoder_InitTypeDef sConfig = {0};
  TIM_MasterConfigTypeDef sMasterConfig = {0};

  /* USER CODE BEGIN TIM1_Init 1 */

  /* USER CODE END TIM1_Init 1 */
  htim1.Instance = TIM1;
  htim1.Init.Prescaler = 0;
  htim1.Init.CounterMode = TIM_COUNTERMODE_UP;
  htim1.Init.Period = 65535;
  htim1.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
  htim1.Init.RepetitionCounter = 0;
  htim1.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_DISABLE;
  sConfig.EncoderMode = TIM_ENCODERMODE_TI1;
  sConfig.IC1Polarity = TIM_ICPOLARITY_RISING;
  sConfig.IC1Selection = TIM_ICSELECTION_DIRECTTI;
  sConfig.IC1Prescaler = TIM_ICPSC_DIV1;
  sConfig.IC1Filter = 0;
  sConfig.IC2Polarity = TIM_ICPOLARITY_RISING;
  sConfig.IC2Selection = TIM_ICSELECTION_DIRECTTI;
  sConfig.IC2Prescaler = TIM_ICPSC_DIV1;
  sConfig.IC2Filter = 0;
  if (HAL_TIM_Encoder_Init(&htim1, &sConfig) != HAL_OK)
  {
    Error_Handler();
  }
  sMasterConfig.MasterOutputTrigger = TIM_TRGO_RESET;
  sMasterConfig.MasterSlaveMode = TIM_MASTERSLAVEMODE_DISABLE;
  if (HAL_TIMEx_MasterConfigSynchronization(&htim1, &sMasterConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN TIM1_Init 2 */

  /* USER CODE END TIM1_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOC_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOC, GPIO_PIN_13|GPIO_PIN_14|GPIO_PIN_15, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOA, GPIO_PIN_0|GPIO_PIN_10|GPIO_PIN_11|GPIO_PIN_12 
                          |GPIO_PIN_15, GPIO_PIN_RESET);

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOB, GPIO_PIN_0|GPIO_PIN_1|GPIO_PIN_13|GPIO_PIN_14 
                          |GPIO_PIN_15|GPIO_PIN_3|GPIO_PIN_4|GPIO_PIN_5 
                          |GPIO_PIN_6|GPIO_PIN_7, GPIO_PIN_RESET);

  /*Configure GPIO pins : PC13 PC14 PC15 */
  GPIO_InitStruct.Pin = GPIO_PIN_13|GPIO_PIN_14|GPIO_PIN_15;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOC, &GPIO_InitStruct);

  /*Configure GPIO pins : PA0 PA10 PA11 PA12 
                           PA15 */
  GPIO_InitStruct.Pin = GPIO_PIN_0|GPIO_PIN_10|GPIO_PIN_11|GPIO_PIN_12 
                          |GPIO_PIN_15;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  /*Configure GPIO pins : PA3 PA4 PA5 PA6 */
  GPIO_InitStruct.Pin = GPIO_PIN_3|GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  /*Configure GPIO pins : PB0 PB1 PB13 PB14 
                           PB15 PB3 PB4 PB5 
                           PB6 PB7 */
  GPIO_InitStruct.Pin = GPIO_PIN_0|GPIO_PIN_1|GPIO_PIN_13|GPIO_PIN_14 
                          |GPIO_PIN_15|GPIO_PIN_3|GPIO_PIN_4|GPIO_PIN_5 
                          |GPIO_PIN_6|GPIO_PIN_7;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /*Configure GPIO pins : PB10 PB11 PB12 */
  GPIO_InitStruct.Pin = GPIO_PIN_10|GPIO_PIN_11|GPIO_PIN_12;
  GPIO_InitStruct.Mode = GPIO_MODE_INPUT;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

}

/* USER CODE BEGIN 4 */

/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  while(1) 
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{ 
  /* USER CODE BEGIN 6 */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */

/************************ (C) COPYRIGHT STMicroelectronics *****END OF FILE****/
